﻿'1、实现72向量初始化——以便按向量访问棋盘上的点
'2、实现72向量按点分组——以便按点访问棋盘上的向量
'3、实现72向量综合比较——从而得到局面的评价分值

Imports System.Runtime.InteropServices
Public Class mVectors
    '72个可成棋向量，用于类内部计算下面的哈希表。
    Private all As New List(Of mVector)(72)
    '表的键是点坐标（integer），而键值是该点所在的各向量（mVector）。
    Public hs As New Hashtable(225)
    '评价函数返回的冲杀信息
    Public lnkinf As New List(Of LinkInfo)
    '初始化与点相关的向量。
    Public Sub New()
        '求所有的向量
        Dim x, y As Integer
        '横向
        For y = 0 To 14
            all.Add(GetVector(0, y, 14, y, 1, 0))
        Next
        '纵向
        For x = 0 To 14
            all.Add(GetVector(x, 0, x, 14, 0, 1))
        Next
        '右上
        For y = 4 To 14
            all.Add(GetVector(0, y, y, 0, 1, -1))
        Next
        For x = 1 To 10
            all.Add(GetVector(x, 14, 14, x, 1, -1))
        Next
        '左上
        For x = 4 To 14
            all.Add(GetVector(x, 14, 0, 14 - x, -1, -1))
        Next
        For y = 13 To 4 Step -1
            all.Add(GetVector(14, y, 14 - y, 0, -1, -1))
        Next
        '分配到点记录表
        Dim i As Integer
        For x = 0 To 14
            For y = 0 To 14
                Dim ls As New List(Of mVector)
                '遍历全部向量，将点所在的向量保存到ls。
                For i = 0 To 71
                    If InLine(i, y * 15 + x) <> -1 Then ls.Add(all(i))
                Next
                '以点坐标为键，加入表中。
                hs.Add(y * 15 + x, ls)
            Next
        Next
    End Sub

    '根据起点终点初始化全部坐标点（用一个字节表示）
    Private Function GetVector(x1 As Integer, y1 As Integer, x2 As Integer, y2 As Integer, dx As Integer, dy As Integer) As mVector
        '向量上的全部点。
        Dim ps() As Byte = Nothing
        '当前坐标X,Y,记数。
        Dim x = -1, y = -1, cst As Integer
        '从向量起点遍历，直到终点，把每一个点记录下来。
        Do Until x = x2 AndAlso y = y2
            x = x1 + dx * cst
            y = y1 + dy * cst
            ReDim Preserve ps(cst)
            ps(cst) = y * 15 + x        '将坐标转换为数组下标
            cst += 1
        Loop
        Return New mVector(ps)
    End Function

    '点是否在这个形体内，若不在为-1。
    Private Function InLine(index As Integer, b As Integer) As Integer
        Return Array.IndexOf(Of Byte)(all(index).ps, b)
    End Function

    '评价局面状态：被杀死、杀死对手、被冲死、冲死对手、被冲、冲对方、一般棋型。
    Public Function Evaluate(ucpc As mBitBoard, player As Integer, RestrictedMove As Integer) As Integer
        '循环变量
        Dim j As Integer
        '冲棋标志
        Dim cq As Boolean
        '清理记录
        lnkinf.Clear()
        '双方的长连、成5、活4、冲4、活3、冲3、活2向量。
        Dim l60_1, l50_1, l42_1, l41_1, l32_1, l31_1, l22_1 As New List(Of LinkInfo)    '本方
        Dim l60_2, l50_2, l42_2, l41_2, l32_2, l31_2, l22_2 As New List(Of LinkInfo)    '对方

        '统计各向量上的棋型
        For j = 0 To 71
            If all(j).update Then       '若需要，则更新棋型评价。
                all(j).Evaluate(ucpc)   '更新函数同时更新黑白两色。
                all(j).update = False
            End If
            '统计本方、对方的72向量上各种棋型的信息。
            '本方
            Select Case all(j).linkInfs(player).lnk
                Case 60                                 '长连
                    l60_1.Add(all(j).linkInfs(player))
                Case 50                                 '成5
                    l50_1.Add(all(j).linkInfs(player))
                Case 42                                 '活4
                    l42_1.Add(all(j).linkInfs(player))
                Case 41                                 '冲4
                    l41_1.Add(all(j).linkInfs(player))
                Case 32                                 '活3
                    l32_1.Add(all(j).linkInfs(player))
                Case 31                                 '冲3
                    l31_1.Add(all(j).linkInfs(player))
                Case 22                                 '活2
                    l22_1.Add(all(j).linkInfs(player))
            End Select
            '对方
            Select Case all(j).linkInfs(1 - player).lnk
                Case 60                                 '长连
                    l60_2.Add(all(j).linkInfs(1 - player))
                Case 50                                 '成5
                    l50_2.Add(all(j).linkInfs(1 - player))
                Case 42                                 '活4
                    l42_2.Add(all(j).linkInfs(1 - player))
                Case 41                                 '冲4
                    l41_2.Add(all(j).linkInfs(1 - player))
                Case 32                                 '活3
                    l32_2.Add(all(j).linkInfs(1 - player))
                Case 31                                 '冲3
                    l31_2.Add(all(j).linkInfs(1 - player))
                Case 22                                 '活2
                    l22_2.Add(all(j).linkInfs(1 - player))
            End Select
        Next
        '====================================死棋棋型（长连、禁手、成5）======================================
        '1、被杀死
        '1.1、对方成5
        If l50_2.Count > 0 Then Return -10000
        '1.2、对方不禁手
        If RestrictedMove <> 1 - player Then
            If l60_2.Count > 0 Then Return -10000
        End If
        '1.3、被禁手
        If RestrictedMove = player Then
            If l60_1.Count > 0 AndAlso l50_1.Count = 0 Then Return -10000 '长连，但是同时成5不为禁手。
            If l32_1.Count > 1 Then Return -10000 '双活三禁手 
            If l42_1.Count > 1 Then Return -10000 '双活四禁手
            If l41_1.Count > 1 Then Return -10000 '双冲四禁手
        End If

        '2、杀死对方
        '2.1、成5
        If l50_1.Count > 0 Then Return 10000
        '2.2、不禁手
        If RestrictedMove <> player Then
            If l60_1.Count > 0 Then Return 10000
        End If
        '2.3、对方禁手
        If RestrictedMove = 1 - player Then
            If l60_2.Count > 0 AndAlso l50_2.Count = 0 Then Return 10000 '长连，但是同时成5不为禁手。
            If l32_2.Count > 1 Then Return 10000 '双活三禁手 
            If l42_2.Count > 1 Then Return 10000 '双活四禁手
            If l41_2.Count > 1 Then Return 10000 '双冲四禁手
        End If

        '==================================可能无解棋（活4、冲4、活3）=====================================
        '不一定具体判断是否有解，无法判断时只需要返回全部可能解，将军延伸会给出正确评价。
        '3、对方冲棋
        '3.1、对方活4，若仅有一个活4，确切解为己方冲4、己方活4的冲点。
        If l42_2.Count > 0 Then
            lnkinf.AddRange(l41_1)
            lnkinf.AddRange(l42_1)
            cq = True
        End If
        '3.2、对方冲4，若仅有一个冲4，确切解为对方冲4、己方冲4、己方活4的冲点。
        If l41_2.Count > 0 Then
            lnkinf.AddRange(l41_2)
            lnkinf.AddRange(l41_1)
            lnkinf.AddRange(l42_1)
            cq = True
        End If
        '3.3、对方活3，若仅有一个活3，确切解为对方活3、己方冲4、己方活4、己方冲3、己方活3的冲点。
        If l32_2.Count > 0 Then
            lnkinf.AddRange(l32_2)
            lnkinf.AddRange(l41_1)
            lnkinf.AddRange(l42_1)
            lnkinf.AddRange(l31_1)
            lnkinf.AddRange(l32_1)
            cq = True
        End If
        If cq Then Return -5000 '对方冲棋

        '4、己方冲棋
        '4.1、己方活4，若仅有一个活4，确切解为对方冲4、对方活4的冲点。
        If l42_1.Count > 0 Then
            lnkinf.AddRange(l41_2)
            lnkinf.AddRange(l42_2)
            cq = True
        End If
        '4.2、己方冲4，若仅有一个冲4，确切解为己方冲4、对方冲4、对方活4的冲点。
        If l41_1.Count > 0 Then
            lnkinf.AddRange(l41_1)
            lnkinf.AddRange(l41_2)
            lnkinf.AddRange(l42_2)
            cq = True
        End If
        '4.3、己方活3，若仅有一个活3，确切解为己方活3、对方冲4、对方活4、对方冲3、对方活3的冲点。
        If l32_1.Count > 0 Then
            lnkinf.AddRange(l32_1)
            lnkinf.AddRange(l41_2)
            lnkinf.AddRange(l42_2)
            lnkinf.AddRange(l31_2)
            lnkinf.AddRange(l32_2)
            cq = True
        End If
        If cq Then Return 5000 '己方冲棋

        '=====================================一般棋型（冲3、活2）======================================
        '5、其他情况，冲3，活2
        Dim pvl1 = l31_1.Count * 20 + l22_1.Count * 30
        Dim pvl2 = l31_2.Count * 20 + l22_2.Count * 30
        Return pvl1 - pvl2
    End Function

End Class